"""Contains the `MeshSlepianNoiseField` class."""
import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._string_methods
import sleplet._validation
import sleplet.meshes.mesh_slepian_field
import sleplet.noise
from sleplet.meshes.mesh_slepian_coefficients import MeshSlepianCoefficients


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class MeshSlepianNoiseField(MeshSlepianCoefficients):
    """
    Create a noisedfield on a given mesh computed from a Slepian region of the
    mesh. The default field is the per-vertex normals of the mesh.
    """

    SNR: float = -5
    """A parameter which controls the level of signal-to-noise in the noised
    data."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        smf = sleplet.meshes.mesh_slepian_field.MeshSlepianField(
            self.mesh,
            region=True,
        )
        noise = sleplet.noise._create_slepian_mesh_noise(
            self.mesh_slepian,
            smf.coefficients,
            self.SNR,
        )
        sleplet.noise.compute_snr(smf.coefficients, noise, "Slepian")
        return noise

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"slepian_{self.mesh.name}_noise_field"
            f"{sleplet._string_methods.filename_args(self.SNR, 'snr')}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return False

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 1
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be {num_args}"
                raise ValueError(msg)
            self.SNR = self.extra_args[0]



import dataclasses
import logging

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._integration_methods
import sleplet._validation
import sleplet._vars
from sleplet.slepian.slepian_functions import SlepianFunctions

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class SlepianDecomposition:
    L: int
    slepian: SlepianFunctions
    _: dataclasses.KW_ONLY
    f: npt.NDArray[np.complex_] | None = None
    flm: npt.NDArray[np.complex_ | np.float_] | None = None
    mask: npt.NDArray[np.float_] | None = None
    _method: str = pydantic.Field(default="", init_var=False, repr=False)

    def __post_init__(self: typing_extensions.Self) -> None:
        self._detect_method()

    def decompose(self: typing_extensions.Self, rank: int) -> complex:
        """Decompose the signal into its Slepian coefficients via the given method."""
        self._validate_rank(rank)

        match self._method:
            case "harmonic_sum":
                return self._harmonic_sum(rank)
            case "integrate_sphere":
                return self._integrate_sphere(rank)
            case "integrate_region":
                return self._integrate_region(rank)
            case _:
                msg = f"'{self._method}' is not a valid method"
                raise ValueError(msg)

    def decompose_all(
        self: typing_extensions.Self,
        n_coefficients: int,
    ) -> npt.NDArray[np.complex_]:
        """Decompose all ranks of the Slepian coefficients."""
        coefficients = np.zeros(n_coefficients, dtype=np.complex_)
        for rank in range(n_coefficients):
            coefficients[rank] = self.decompose(rank)
        return coefficients

    def _integrate_region(self: typing_extensions.Self, rank: int) -> complex:
        r"""
        F_{p} =
        \frac{1}{\lambda_{p}}
        \int\limits_{R} \dd{\Omega(\omega)}
        f(\omega) \overline{S_{p}(\omega)}.
        """
        s_p = ssht.inverse(
            self.slepian.eigenvectors[rank],
            self.L,
            Method=sleplet._vars.SAMPLING_SCHEME,
        )
        weight = sleplet._integration_methods.calc_integration_weight(self.L)
        integration = sleplet._integration_methods.integrate_region_sphere(
            self.mask,
            weight,
            self.f,
            s_p.conj(),
        )
        return integration / self.slepian.eigenvalues[rank]

    def _integrate_sphere(self: typing_extensions.Self, rank: int) -> complex:
        r"""
        F_{p} =
        \int\limits_{S^{2}} \dd{\Omega(\omega)}
        f(\omega) \overline{S_{p}(\omega)}.
        """
        s_p = ssht.inverse(
            self.slepian.eigenvectors[rank],
            self.L,
            Method=sleplet._vars.SAMPLING_SCHEME,
        )
        weight = sleplet._integration_methods.calc_integration_weight(self.L)
        return sleplet._integration_methods.integrate_whole_sphere(
            weight,
            self.f,
            s_p.conj(),
        )

    def _harmonic_sum(self: typing_extensions.Self, rank: int) -> complex:
        r"""
        F_{p} =
        \sum\limits_{\ell=0}^{L^{2}}
        \sum\limits_{m=-\ell}^{\ell}
        f_{\ell m} (S_{p})_{\ell m}^{*}.
        """
        return (self.flm * self.slepian.eigenvectors[rank].conj()).sum()

    def _detect_method(self: typing_extensions.Self) -> None:
        """Detect what method is used to perform the decomposition."""
        if self.flm is not None:
            _logger.info("harmonic sum method selected")
            self._method = "harmonic_sum"
        elif self.f is not None:
            if self.mask is None:
                _logger.info("integrating the whole sphere method selected")
                self._method = "integrate_sphere"
            else:
                _logger.info("integrating a region on the sphere method selected")
                self._method = "integrate_region"
        else:
            msg = (
                "need to pass one off harmonic coefficients, real pixels "
                "or real pixels with a mask"
            )
            raise RuntimeError(msg)

    def _validate_rank(self: typing_extensions.Self, rank: int) -> None:
        """Check the requested rank is valid."""
        if not isinstance(rank, int):
            msg = "rank should be an integer"
            raise TypeError(msg)
        if rank < 0:
            msg = "rank cannot be negative"
            raise ValueError(msg)
        if rank >= self.L**2:
            msg = f"rank should be less than {self.L**2}"
            raise ValueError(msg)



"""Contains the abstract `Flm` class."""
import abc

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._validation
import sleplet.noise
from sleplet.functions.coefficients import Coefficients


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class Flm(Coefficients):
    """Abstract parent class to handle harmonic coefficients on the sphere."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def rotate(  # noqa: D102
        self: typing_extensions.Self,
        alpha: float,
        beta: float,
        *,
        gamma: float = 0,
    ) -> npt.NDArray[np.complex_]:
        return ssht.rotate_flms(self.coefficients, alpha, beta, gamma, self.L)

    def _translation_helper(
        self: typing_extensions.Self,
        alpha: float,
        beta: float,
    ) -> npt.NDArray[np.complex_]:
        return ssht.create_ylm(beta, alpha, self.L).conj().flatten()

    def _add_noise_to_signal(
        self: typing_extensions.Self,
    ) -> tuple[npt.NDArray[np.complex_ | np.float_] | None, float | None]:
        """Add Gaussian white noise to the signal."""
        self.coefficients: npt.NDArray[np.complex_ | np.float_]
        if self.noise is not None:
            unnoised_coefficients = self.coefficients.copy()
            nlm = sleplet.noise._create_noise(self.L, self.coefficients, self.noise)
            snr = sleplet.noise.compute_snr(self.coefficients, nlm, "Harmonic")
            self.coefficients = self.coefficients + nlm
            return unnoised_coefficients, snr
        return None, None

    @abc.abstractmethod
    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        raise NotImplementedError

    @abc.abstractmethod
    def _create_name(self: typing_extensions.Self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def _set_reality(self: typing_extensions.Self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def _set_spin(self: typing_extensions.Self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def _setup_args(self: typing_extensions.Self) -> None:
        raise NotImplementedError



import logging
import os

import platformdirs
import pooch

_logger = logging.getLogger(__name__)


_ZENODO_DATA_DOI = "10.5281/zenodo.7767698"
_POOCH = pooch.create(
    path=pooch.os_cache("sleplet"),
    base_url=f"doi:{_ZENODO_DATA_DOI}/",
    registry=None,
)
_POOCH.load_registry_from_doi()


def find_on_pooch_then_local(filename: str) -> os.PathLike[str] | None:
    """Find a file on POOCH first and if not look in data folder."""
    if filename in _POOCH.registry:
        msg = f"Found {filename} at https://doi.org/{_ZENODO_DATA_DOI}"
        _logger.info(msg)
        return _POOCH.fetch(filename, progressbar=True)
    if (platformdirs.user_data_path() / filename).exists():
        msg = f"Found {filename} at {platformdirs.user_data_path() / filename}"
        _logger.info(msg)
        return platformdirs.user_data_path() / filename
    msg = f"No {filename} found, calculating..."
    _logger.info(msg)
    return None



import logging

import numpy as np
import numpy.typing as npt

import pyssht as ssht

_logger = logging.getLogger(__name__)


def apply_gaussian_smoothing(
    flm: npt.NDArray[np.complex_],
    L: int,
    smoothing_factor: int,
) -> npt.NDArray[np.complex_]:
    """
    Apply Gaussian smoothing to the given signal.

    s_lm = exp(-ell^2 sigma^2)
    s(omega) = exp(-theta^2 / sigma^2)
    FWHM = 2 * sqrt(ln2) * sigma
    """
    sigma = np.pi / (smoothing_factor * L)
    fwhm = 2 * np.sqrt(np.log(2)) * sigma
    msg = f"FWHM = {np.rad2deg(fwhm):.2f}degrees"
    _logger.info(msg)
    return ssht.gaussian_smoothing(flm, L, sigma)



import fractions
import re
import typing

import numpy as np

if typing.TYPE_CHECKING:
    import sleplet.functions.coefficients
    import sleplet.meshes.mesh_coefficients


def _get_angle_num_dem(angle_fraction: float) -> tuple[int, int]:
    """Get numerator and denominator for a given decimal."""
    angle = fractions.Fraction(angle_fraction).limit_denominator()
    return angle.numerator, angle.denominator


def _pi_in_filename(numerator: int, denominator: int) -> str:
    """Create filename for angle as multiple of pi."""
    filename = f"{numerator}pi"
    # if whole number
    if denominator != 1:
        filename += f"{denominator}"
    return filename


def filename_args(value: float, arg_name: str) -> str:
    """Add an extra argument to filename."""
    filename = "_"
    num, dem = _get_angle_num_dem(value)
    filename += f"{num}{arg_name}"
    if abs(value) < 1 and value != 0:
        filename += f"{dem}"
    return filename


def filename_angle(
    alpha_pi_fraction: float,
    beta_pi_fraction: float,
    gamma_pi_fraction: float = 0,
) -> str:
    """Middle part of filename."""
    # get numerator/denominator for filename
    alpha_num, alpha_den = _get_angle_num_dem(alpha_pi_fraction)
    beta_num, beta_den = _get_angle_num_dem(beta_pi_fraction)
    gamma_num, gamma_den = _get_angle_num_dem(gamma_pi_fraction)

    match (alpha_num, beta_num):
        case (0, 0):
            filename = "alpha0_beta0"
        case (0, _):
            filename = f"alpha0_beta{_pi_in_filename(beta_num, beta_den)}"
        case (_, 0):
            filename = f"alpha{_pi_in_filename(alpha_num, alpha_den)}_beta0"
        case _:
            filename = (
                f"alpha{_pi_in_filename(alpha_num, alpha_den)}"
                f"_beta{_pi_in_filename(beta_num, beta_den)}"
            )

    # if rotation with gamma != 0
    if gamma_num:
        filename += f"_gamma{_pi_in_filename(gamma_num, gamma_den)}"
    return filename


def multiples_of_pi(angle: float) -> str:
    """
    Print the unicode pi with a prefix of the multiple unless it's 1
    i.e. pi, 2pi, 3pi.
    """
    multiple = int(angle / np.pi)
    return f"{multiple if multiple != 1 else ''}\u03C0"


def angle_as_degree(radian: float) -> int:
    """Convert radian angle to integer degree."""
    return round(np.rad2deg(radian))


def wavelet_ending(j_min: int, j: int | None) -> str:
    """Create the ending name of the given wavelet."""
    return "_scaling" if j is None else f"{filename_args(j + j_min, 'j')}"


def _convert_camel_case_to_snake_case(name: str) -> str:
    """Convert a string in camel case to snake case."""
    return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()


def convert_classes_list_to_snake_case(
    classes: list[
        type["sleplet.functions.coefficients.Coefficients"]
        | type["sleplet.meshes.mesh_coefficients.MeshCoefficients"]
    ],
    *,
    word_to_remove: str = "",
) -> list[str]:
    """Convert a list of classes to snake case."""
    return [
        _convert_camel_case_to_snake_case(c.__name__.removeprefix(word_to_remove))
        for c in classes
    ]



import argparse
import logging

import sleplet._class_lists
import sleplet._string_methods
import sleplet.meshes.mesh
import sleplet.meshes.mesh_coefficients
import sleplet.plot_methods
import sleplet.plotting._create_plot_mesh

_logger = logging.getLogger(__name__)


def valid_meshes(mesh_name: str) -> str:
    """Check if valid mesh name."""
    if mesh_name in sleplet._class_lists.MESHES:
        return mesh_name
    msg = f"'{mesh_name}' is not a valid mesh name to plot"
    raise ValueError(msg)


def valid_methods(method_name: str) -> str:
    """Check if valid mesh name."""
    if method_name in sleplet._string_methods.convert_classes_list_to_snake_case(
        sleplet._class_lists.MESH_COEFFICIENTS,
        word_to_remove="Mesh",
    ):
        return method_name
    msg = f"'{method_name}' is not a valid method to plot"
    raise ValueError(msg)


def read_args() -> argparse.Namespace:
    """Read args from the command line."""
    parser = argparse.ArgumentParser(description="Create mesh plot")
    parser.add_argument(
        "function",
        type=valid_meshes,
        choices=sleplet._class_lists.MESHES,
        help="mesh to plot",
    )
    parser.add_argument(
        "--extra_args",
        "-e",
        type=int,
        nargs="+",
        help="list of extra args for functions",
    )
    parser.add_argument(
        "--method",
        "-m",
        type=valid_methods,
        nargs="?",
        default="basis_functions",
        const="basis_functions",
        choices=sleplet._string_methods.convert_classes_list_to_snake_case(
            sleplet._class_lists.MESH_COEFFICIENTS,
            word_to_remove="Mesh",
        ),
        help="plotting routine: defaults to basis",
    )
    parser.add_argument("--noise", "-n", type=int, help="the SNR_IN of the noise level")
    parser.add_argument(
        "--region",
        "-r",
        action="store_true",
        help="flag which masks the function for a region",
    )
    parser.add_argument(
        "--unnormalise",
        "-u",
        action="store_true",
        help="flag turns off normalisation for plot",
    )
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=sleplet.__version__,
    )
    parser.add_argument(
        "--zoom",
        "-z",
        action="store_true",
        help="flag which zooms in on the region of interest",
    )
    return parser.parse_args()


def plot(
    f: sleplet.meshes.mesh_coefficients.MeshCoefficients,
    *,
    normalise: bool,
    amplitude: float | None,
) -> None:
    """Master plotting method."""
    field = sleplet.plot_methods._coefficients_to_field_mesh(f, f.coefficients)
    sleplet.plotting._create_plot_mesh.PlotMesh(
        f.mesh,
        f.name,
        field,
        amplitude=amplitude,
        normalise=normalise,
        region=hasattr(f, "mesh_slepian"),
    ).execute()


def main() -> None:
    args = read_args()
    msg = f"mesh: '{args.function}', plotting method: '{args.method}'"
    _logger.info(msg)

    # function to plot
    mesh = sleplet.meshes.mesh.Mesh(args.function, zoom=args.zoom)
    f = sleplet._class_lists.MESH_COEFFICIENTS[
        sleplet._string_methods.convert_classes_list_to_snake_case(
            sleplet._class_lists.MESH_COEFFICIENTS,
            word_to_remove="Mesh",
        ).index(args.method)
    ](
        mesh,
        extra_args=args.extra_args,
        noise=args.noise if args.noise is not None else None,
        region=args.region,
    )

    # custom amplitude for noisy plots
    amplitude = sleplet.plot_methods.compute_amplitude_for_noisy_mesh_plots(f)

    # perform plot
    plot(f, normalise=not args.unnormalise, amplitude=amplitude)


if __name__ == "__main__":
    main()



"""Set of classes to create functions on the sphere."""

from .africa import Africa
from .axisymmetric_wavelet_coefficients_africa import (
    AxisymmetricWaveletCoefficientsAfrica,
)
from .axisymmetric_wavelet_coefficients_earth import (
    AxisymmetricWaveletCoefficientsEarth,
)
from .axisymmetric_wavelet_coefficients_south_america import (
    AxisymmetricWaveletCoefficientsSouthAmerica,
)
from .axisymmetric_wavelets import AxisymmetricWavelets
from .dirac_delta import DiracDelta
from .earth import Earth
from .elongated_gaussian import ElongatedGaussian
from .gaussian import Gaussian
from .harmonic_gaussian import HarmonicGaussian
from .identity import Identity
from .noise_earth import NoiseEarth
from .ridgelets import Ridgelets
from .slepian import Slepian
from .slepian_africa import SlepianAfrica
from .slepian_dirac_delta import SlepianDiracDelta
from .slepian_identity import SlepianIdentity
from .slepian_noise_africa import SlepianNoiseAfrica
from .slepian_noise_south_america import SlepianNoiseSouthAmerica
from .slepian_south_america import SlepianSouthAmerica
from .slepian_wavelet_coefficients_africa import SlepianWaveletCoefficientsAfrica
from .slepian_wavelet_coefficients_south_america import (
    SlepianWaveletCoefficientsSouthAmerica,
)
from .slepian_wavelets import SlepianWavelets
from .south_america import SouthAmerica
from .spherical_harmonic import SphericalHarmonic
from .squashed_gaussian import SquashedGaussian
from .wmap import Wmap

__all__ = [
    "Africa",
    "AxisymmetricWaveletCoefficientsAfrica",
    "AxisymmetricWaveletCoefficientsEarth",
    "AxisymmetricWaveletCoefficientsSouthAmerica",
    "AxisymmetricWavelets",
    "DiracDelta",
    "Earth",
    "ElongatedGaussian",
    "Gaussian",
    "HarmonicGaussian",
    "Identity",
    "NoiseEarth",
    "Ridgelets",
    "Slepian",
    "SlepianAfrica",
    "SlepianDiracDelta",
    "SlepianIdentity",
    "SlepianNoiseAfrica",
    "SlepianNoiseSouthAmerica",
    "SlepianSouthAmerica",
    "SlepianWaveletCoefficientsAfrica",
    "SlepianWaveletCoefficientsSouthAmerica",
    "SlepianWavelets",
    "SouthAmerica",
    "SphericalHarmonic",
    "SquashedGaussian",
    "Wmap",
]



"""Contains the `SlepianSouthAmerica` class."""
import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._string_methods
import sleplet._validation
import sleplet.functions.south_america
import sleplet.slepian.region
from sleplet.functions.fp import Fp


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class SlepianSouthAmerica(Fp):
    """
    Create a Slepian region on the topographic map of the Earth of the
    South America region.
    """

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()
        if (
            isinstance(self.region, sleplet.slepian.region.Region)
            and self.region._name_ending != "south_america"
        ):
            msg = "Slepian region selected must be 'south_america'"
            raise RuntimeError(msg)

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        sa = sleplet.functions.south_america.SouthAmerica(
            self.L,
            smoothing=self.smoothing,
        )
        return sleplet.slepian_methods.slepian_forward(
            self.L,
            self.slepian,
            flm=sa.coefficients,
        )

    def _create_name(self: typing_extensions.Self) -> str:
        return sleplet._string_methods._convert_camel_case_to_snake_case(
            self.__class__.__name__,
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return False

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            msg = f"{self.__class__.__name__} does not support extra arguments"
            raise TypeError(msg)



import dataclasses
import logging

import cmocean
import matplotlib as mpl
import numpy as np
import numpy.typing as npt
import plotly.graph_objs as go
import plotly.io as pio
import pydantic
import typing_extensions

import sleplet._mask_methods
import sleplet._mesh_methods
import sleplet._plotly_methods
import sleplet._validation
import sleplet.meshes.mesh
import sleplet.plot_methods

_logger = logging.getLogger(__name__)

_MESH_CBAR_LEN = 0.95
_MESH_CBAR_FONT_SIZE = 32
_MESH_UNSEEN = -1e5  # kaleido bug


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class PlotMesh:
    """Create surface mesh plot via `plotly`."""

    mesh: sleplet.meshes.mesh.Mesh
    """The given mesh object."""
    filename: str
    """The output filename of the plot."""
    f: npt.NDArray[np.complex_ | np.float_]
    """The field value sampled on the mesh."""
    _: dataclasses.KW_ONLY
    amplitude: float | None = None
    """Whether to customise the amplitude range of the colour bar."""
    normalise: bool = True
    """Whether to normalise the plot or not."""
    region: bool = False
    """Whether to set the field values outside of the region to zero."""

    def __post_init__(self: typing_extensions.Self) -> None:
        if self.normalise:
            self.filename += "_norm"

    def execute(
        self: typing_extensions.Self,
        colour: mpl.colors.LinearSegmentedColormap = cmocean.cm.ice,
    ) -> None:
        """
        Perform the plot.

        Args:
            colour: From the `cmocean.cm` module
        """
        vmin, vmax = self.f.min(), self.f.max()
        f = self._prepare_field(self.f)

        if self.region:
            # make plot area clearer
            f = self._set_outside_region_to_minimum(f)

        # pick largest tick max value
        tick_mark = sleplet._plotly_methods.create_tick_mark(
            vmin,
            vmax,
            amplitude=self.amplitude,
        )

        data = [
            go.Mesh3d(
                x=self.mesh.vertices[:, 0],
                y=self.mesh.vertices[:, 2],
                z=self.mesh.vertices[:, 1],
                i=self.mesh.faces[:, 0],
                j=self.mesh.faces[:, 1],
                k=self.mesh.faces[:, 2],
                intensitymode="cell",
                intensity=f,
                cmax=1 if self.normalise else tick_mark,
                cmid=0.5 if self.normalise else 0,
                cmin=0 if self.normalise else -tick_mark,
                colorbar=sleplet._plotly_methods.create_colour_bar(
                    tick_mark,
                    normalise=self.normalise,
                    bar_len=_MESH_CBAR_LEN,
                    bar_pos=self.mesh._colourbar_pos,
                    font_size=_MESH_CBAR_FONT_SIZE,
                ),
                colorscale=sleplet.plot_methods._convert_colourscale(colour),
                lighting=go.mesh3d.Lighting(ambient=1),
                reversescale=True,
            ),
        ]

        layout = sleplet._plotly_methods.create_layout(self.mesh._camera_view)

        fig = go.Figure(data=data, layout=layout)

        msg = f"Opening: {self.filename}"
        _logger.info(msg)
        pio.show(fig, config={"toImageButtonOptions": {"filename": self.filename}})

    def _prepare_field(
        self: typing_extensions.Self,
        f: npt.NDArray[np.complex_ | np.float_],
    ) -> npt.NDArray[np.float_]:
        """Scales the field before plotting."""
        return sleplet.plot_methods._normalise_function(
            sleplet._mesh_methods.average_functions_on_vertices_to_faces(
                self.mesh.faces,
                f,
            ),
            normalise=self.normalise,
        )

    def _set_outside_region_to_minimum(
        self: typing_extensions.Self,
        f: npt.NDArray[np.float_],
    ) -> npt.NDArray[np.float_]:
        """
        For the Slepian region set the outside area to negative infinity
        hence it is clear we are only interested in the coloured region.
        """
        region_on_faces = sleplet._mask_methods.convert_region_on_vertices_to_faces(
            self.mesh,
        )
        return np.where(region_on_faces, f, _MESH_UNSEEN)



import pathlib

import sleplet.functions.africa
import sleplet.functions.axisymmetric_wavelet_coefficients_africa
import sleplet.functions.axisymmetric_wavelet_coefficients_earth
import sleplet.functions.axisymmetric_wavelet_coefficients_south_america
import sleplet.functions.axisymmetric_wavelets
import sleplet.functions.coefficients
import sleplet.functions.dirac_delta
import sleplet.functions.earth
import sleplet.functions.elongated_gaussian
import sleplet.functions.gaussian
import sleplet.functions.harmonic_gaussian
import sleplet.functions.identity
import sleplet.functions.noise_earth
import sleplet.functions.ridgelets
import sleplet.functions.slepian
import sleplet.functions.slepian_africa
import sleplet.functions.slepian_dirac_delta
import sleplet.functions.slepian_identity
import sleplet.functions.slepian_noise_africa
import sleplet.functions.slepian_noise_south_america
import sleplet.functions.slepian_south_america
import sleplet.functions.slepian_wavelet_coefficients_africa
import sleplet.functions.slepian_wavelet_coefficients_south_america
import sleplet.functions.slepian_wavelets
import sleplet.functions.south_america
import sleplet.functions.spherical_harmonic
import sleplet.functions.squashed_gaussian
import sleplet.functions.wmap
import sleplet.meshes.mesh_basis_functions
import sleplet.meshes.mesh_coefficients
import sleplet.meshes.mesh_field
import sleplet.meshes.mesh_noise_field
import sleplet.meshes.mesh_slepian_field
import sleplet.meshes.mesh_slepian_functions
import sleplet.meshes.mesh_slepian_noise_field
import sleplet.meshes.mesh_slepian_wavelet_coefficients
import sleplet.meshes.mesh_slepian_wavelets

_data_path = pathlib.Path(__file__).resolve().parent / "_data"
MESHES: list[str] = [
    pathlib.Path(x).stem.removeprefix("meshes_regions_")
    for x in pathlib.Path(_data_path).glob("*.toml")
]

FLM: list[type[sleplet.functions.coefficients.Coefficients]] = [
    sleplet.functions.africa.Africa,
    sleplet.functions.axisymmetric_wavelet_coefficients_africa.AxisymmetricWaveletCoefficientsAfrica,
    sleplet.functions.axisymmetric_wavelet_coefficients_earth.AxisymmetricWaveletCoefficientsEarth,
    sleplet.functions.axisymmetric_wavelet_coefficients_south_america.AxisymmetricWaveletCoefficientsSouthAmerica,
    sleplet.functions.axisymmetric_wavelets.AxisymmetricWavelets,
    sleplet.functions.dirac_delta.DiracDelta,
    sleplet.functions.earth.Earth,
    sleplet.functions.elongated_gaussian.ElongatedGaussian,
    sleplet.functions.gaussian.Gaussian,
    sleplet.functions.harmonic_gaussian.HarmonicGaussian,
    sleplet.functions.identity.Identity,
    sleplet.functions.noise_earth.NoiseEarth,
    sleplet.functions.ridgelets.Ridgelets,
    sleplet.functions.south_america.SouthAmerica,
    sleplet.functions.spherical_harmonic.SphericalHarmonic,
    sleplet.functions.squashed_gaussian.SquashedGaussian,
    sleplet.functions.wmap.Wmap,
]

FP: list[type[sleplet.functions.coefficients.Coefficients]] = [
    sleplet.functions.slepian.Slepian,
    sleplet.functions.slepian_africa.SlepianAfrica,
    sleplet.functions.slepian_dirac_delta.SlepianDiracDelta,
    sleplet.functions.slepian_identity.SlepianIdentity,
    sleplet.functions.slepian_noise_africa.SlepianNoiseAfrica,
    sleplet.functions.slepian_noise_south_america.SlepianNoiseSouthAmerica,
    sleplet.functions.slepian_south_america.SlepianSouthAmerica,
    sleplet.functions.slepian_wavelet_coefficients_africa.SlepianWaveletCoefficientsAfrica,
    sleplet.functions.slepian_wavelet_coefficients_south_america.SlepianWaveletCoefficientsSouthAmerica,
    sleplet.functions.slepian_wavelets.SlepianWavelets,
]

COEFFICIENTS: list[type[sleplet.functions.coefficients.Coefficients]] = FLM + FP

MAPS_LM: list[type[sleplet.functions.coefficients.Coefficients]] = [
    sleplet.functions.earth.Earth,
    sleplet.functions.south_america.SouthAmerica,
    sleplet.functions.wmap.Wmap,
]

MESH_HARMONIC: list[type[sleplet.meshes.mesh_coefficients.MeshCoefficients]] = [
    sleplet.meshes.mesh_basis_functions.MeshBasisFunctions,
    sleplet.meshes.mesh_field.MeshField,
    sleplet.meshes.mesh_noise_field.MeshNoiseField,
]

MESH_SLEPIAN: list[type[sleplet.meshes.mesh_coefficients.MeshCoefficients]] = [
    sleplet.meshes.mesh_slepian_field.MeshSlepianField,
    sleplet.meshes.mesh_slepian_functions.MeshSlepianFunctions,
    sleplet.meshes.mesh_slepian_noise_field.MeshSlepianNoiseField,
    sleplet.meshes.mesh_slepian_wavelet_coefficients.MeshSlepianWaveletCoefficients,
    sleplet.meshes.mesh_slepian_wavelets.MeshSlepianWavelets,
]

MESH_COEFFICIENTS: list[type[sleplet.meshes.mesh_coefficients.MeshCoefficients]] = (
    MESH_HARMONIC + MESH_SLEPIAN
)



import multiprocess.shared_memory
import numpy as np
import numpy.typing as npt


def split_arr_into_chunks(arr_max: int, ncpu: int) -> list[npt.NDArray[np.int_]]:
    """Split L into a list of arrays for parallelism."""
    arr = np.arange(arr_max)
    size = len(arr)
    arr[size // 2 : size] = arr[size // 2 : size][::-1]
    return [np.sort(arr[i::ncpu]) for i in range(ncpu)]


def create_shared_memory_array(
    array: npt.NDArray[np.float_],
) -> tuple[npt.NDArray[np.float_], multiprocess.shared_memory.SharedMemory]:
    """Create a shared memory array to be used in a parallel function."""
    ext_shared_memory = multiprocess.shared_memory.SharedMemory(
        create=True,
        size=array.nbytes,
    )
    array_ext: npt.NDArray[np.float_] = np.ndarray(
        array.shape,
        dtype=array.dtype,
        buffer=ext_shared_memory.buf,
    )
    return array_ext, ext_shared_memory


def attach_to_shared_memory_block(
    array: npt.NDArray[np.float_],
    ext_shared_memory: multiprocess.shared_memory.SharedMemory,
) -> tuple[npt.NDArray[np.float_], multiprocess.shared_memory.SharedMemory]:
    """Attach an array to the shared memory within the parallel function."""
    int_shared_memory = multiprocess.shared_memory.SharedMemory(
        name=ext_shared_memory.name,
    )
    array_int: npt.NDArray[np.float_] = np.ndarray(
        array.shape,
        dtype=array.dtype,
        buffer=int_shared_memory.buf,
    )
    return array_int, int_shared_memory


def free_shared_memory(*shared_memory: multiprocess.shared_memory.SharedMemory) -> None:
    """Close the shared memory object."""
    for shm in shared_memory:
        shm.close()


def release_shared_memory(
    *shared_memory: multiprocess.shared_memory.SharedMemory,
) -> None:
    """Releases the shared memory object."""
    for shm in shared_memory:
        shm.unlink()



import argparse
import logging

import numpy as np
import numpy.typing as npt

import pyssht as ssht

import sleplet._class_lists
import sleplet._mask_methods
import sleplet._string_methods
import sleplet.functions.coefficients
import sleplet.harmonic_methods
import sleplet.plot_methods
import sleplet.plotting._create_plot_sphere
import sleplet.slepian_methods

_logger = logging.getLogger(__name__)


_ALPHA_DEFAULT = 0.75
_ANNOTATION_COLOUR = "gold"
_ARROW_STYLE = {
    "arrowhead": 0,
    "arrowside": "start",
    "ax": 4,
    "ay": 4,
    "startarrowsize": 0.5,
    "startarrowhead": 6,
}
_BETA_DEFAULT = 0.125


def valid_maps(map_name: str) -> str:
    """Check if valid map."""
    if map_name in sleplet._string_methods.convert_classes_list_to_snake_case(
        sleplet._class_lists.MAPS_LM,
    ):
        return map_name
    msg = f"{map_name} is not a valid map to convolve"
    raise ValueError(msg)


def valid_plotting(func_name: str) -> str:
    """Check if valid function."""
    if func_name in sleplet._string_methods.convert_classes_list_to_snake_case(
        sleplet._class_lists.COEFFICIENTS,
    ):
        return func_name
    msg = f"{func_name} is not a valid function to plot"
    raise ValueError(msg)


def read_args() -> argparse.Namespace:
    """Read args from the command line."""
    parser = argparse.ArgumentParser(description="Create SSHT plot")
    parser.add_argument(
        "function",
        type=valid_plotting,
        choices=sleplet._string_methods.convert_classes_list_to_snake_case(
            sleplet._class_lists.COEFFICIENTS,
        ),
        help="function to plot on the sphere",
    )
    parser.add_argument(
        "--alpha",
        "-a",
        type=float,
        default=_ALPHA_DEFAULT,
        help=f"alpha/phi pi fraction - defaults to {_ALPHA_DEFAULT}",
    )
    parser.add_argument("--bandlimit", "-L", type=int, default=16, help="bandlimit")
    parser.add_argument(
        "--beta",
        "-b",
        type=float,
        default=_BETA_DEFAULT,
        help=f"beta/theta pi fraction - defaults to {_BETA_DEFAULT}",
    )
    parser.add_argument(
        "--convolve",
        "-c",
        type=valid_maps,
        default=None,
        choices=sleplet._string_methods.convert_classes_list_to_snake_case(
            sleplet._class_lists.MAPS_LM,
        ),
        help="glm to perform sifting convolution with i.e. flm x glm*",
    )
    parser.add_argument(
        "--extra_args",
        "-e",
        type=int,
        nargs="+",
        help="list of extra args for functions",
    )
    parser.add_argument(
        "--gamma",
        "-g",
        type=float,
        default=0,
        help="gamma pi fraction - defaults to 0 - rotation only",
    )
    parser.add_argument(
        "--method",
        "-m",
        type=str,
        nargs="?",
        default="north",
        const="north",
        choices=["north", "rotate", "translate"],
        help="plotting routine: defaults to north",
    )
    parser.add_argument("--noise", "-n", type=int, help="the SNR_IN of the noise level")
    parser.add_argument(
        "--outline",
        "-o",
        action="store_false",
        help="flag which removes any annotation",
    )
    parser.add_argument(
        "--perspective",
        "-p",
        type=str,
        nargs="?",
        default="south_america",
        const="south_america",
        choices=["africa", "south_america"],
        help="view of Earth: defaults to 'south_america'",
    )
    parser.add_argument(
        "--region",
        "-r",
        action="store_true",
        help="flag which masks the function for a region",
    )
    parser.add_argument(
        "--smoothing",
        "-s",
        type=int,
        help="the scaling of the sigma in Gaussian smoothing of the Earth",
    )
    parser.add_argument(
        "--type",
        "-t",
        type=str,
        nargs="?",
        default="real",
        const="real",
        choices=["abs", "real", "imag", "sum"],
        help="plotting type: defaults to real",
    )
    parser.add_argument(
        "--unnormalise",
        "-u",
        action="store_true",
        help="flag turns off normalisation for plot",
    )
    parser.add_argument(
        "--unzeropad",
        "-z",
        action="store_true",
        help="flag turns off upsampling for plot",
    )
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=sleplet.__version__,
    )
    return parser.parse_args()


def plot(  # noqa: PLR0913
    f: sleplet.functions.coefficients.Coefficients,
    g: sleplet.functions.coefficients.Coefficients | None,
    *,
    alpha_pi_frac: float,
    beta_pi_frac: float,
    gamma_pi_frac: float,
    annotations: bool,
    normalise: bool,
    method: str,
    plot_type: str,
    upsample: bool,
    earth_view: str,
    amplitude: float | None,
) -> None:
    """Master plotting method."""
    filename = f.name
    coefficients = f.coefficients

    # turn off annotation if needed
    msg = f"annotations on: {annotations}"
    _logger.info(msg)
    annotation = []

    # Shannon number for Slepian coefficients
    shannon = f.slepian.N if hasattr(f, "slepian") else None

    msg = f"plotting method: '{method}'"
    _logger.info(msg)
    match method:
        case "rotate":
            coefficients, filename = _rotation_helper(
                f,
                filename,
                alpha_pi_frac,
                beta_pi_frac,
                gamma_pi_frac,
            )
        case "translate":
            coefficients, filename, trans_annotation = _translation_helper(
                f,
                filename,
                alpha_pi_frac,
                beta_pi_frac,
                shannon,
            )

            # annotate translation point
            if annotations:
                annotation.append(trans_annotation)

    if g is not None:
        coefficients, filename = _convolution_helper(
            f,
            g,
            coefficients,
            shannon,
            filename,
        )

    # rotate plot of Earth
    if "earth" in filename:
        match earth_view:
            case "africa":
                coefficients = sleplet.harmonic_methods.rotate_earth_to_africa(
                    coefficients,
                    f.L,
                )
                filename += "_africa"
            case "south_america":
                coefficients = sleplet.harmonic_methods.rotate_earth_to_south_america(
                    coefficients,
                    f.L,
                )

    # get field value
    field = sleplet.plot_methods._coefficients_to_field_sphere(f, coefficients)

    # do plot
    sleplet.plotting._create_plot_sphere.PlotSphere(
        field,
        f.L,
        filename,
        amplitude=amplitude,
        annotations=annotation,
        normalise=normalise,
        plot_type=plot_type,
        reality=f.reality,
        region=f.region if hasattr(f, "slepian") else None,
        spin=f.spin,
        upsample=upsample,
    ).execute()


def _rotation_helper(
    f: sleplet.functions.coefficients.Coefficients,
    filename: str,
    alpha_pi_frac: float,
    beta_pi_frac: float,
    gamma_pi_frac: float,
) -> tuple[npt.NDArray[np.complex_], str]:
    """Perform the rotation specific steps."""
    msg = (
        "angles: (alpha, beta, gamma) = "
        f"({alpha_pi_frac}, {beta_pi_frac}, {gamma_pi_frac})",
    )
    _logger.info(msg)
    filename += (
        "_rotate_"
        f"{sleplet._string_methods.filename_angle(alpha_pi_frac, beta_pi_frac, gamma_pi_frac)}"  # noqa: E501
    )

    # calculate angles
    alpha, beta = sleplet.plot_methods._calc_nearest_grid_point(
        f.L,
        alpha_pi_frac,
        beta_pi_frac,
    )
    gamma = gamma_pi_frac * np.pi

    # rotate by alpha, beta, gamma
    coefficients = f.rotate(alpha, beta, gamma=gamma)
    return coefficients, filename


def _translation_helper(
    f: sleplet.functions.coefficients.Coefficients,
    filename: str,
    alpha_pi_frac: float,
    beta_pi_frac: float,
    shannon: int | None,
) -> tuple[npt.NDArray[np.complex_ | np.float_], str, dict[str, float | int]]:
    """Perform the translation specific steps."""
    msg = f"angles: (alpha, beta) = ({alpha_pi_frac}, {beta_pi_frac})"
    _logger.info(msg)
    # don't add gamma if translation
    filename += (
        "_translate_"
        f"{sleplet._string_methods.filename_angle(alpha_pi_frac, beta_pi_frac)}"
    )

    # calculate angles
    alpha, beta = sleplet.plot_methods._calc_nearest_grid_point(
        f.L,
        alpha_pi_frac,
        beta_pi_frac,
    )

    # translate by alpha, beta
    coefficients = f.translate(alpha, beta, shannon=shannon)

    # annotate translation point
    x, y, z = ssht.s2_to_cart(beta, alpha)
    annotation = {
        "x": x,
        "y": y,
        "z": z,
        "arrowcolor": _ANNOTATION_COLOUR,
    } | _ARROW_STYLE
    return coefficients, filename, annotation


def _convolution_helper(
    f: sleplet.functions.coefficients.Coefficients,
    g: sleplet.functions.coefficients.Coefficients,
    coefficients: npt.NDArray[np.complex_ | np.float_],
    shannon: int | None,
    filename: str,
) -> tuple[npt.NDArray[np.complex_ | np.float_], str]:
    """Perform the convolution specific steps."""
    g_coefficients = (
        sleplet.slepian_methods.slepian_forward(f.L, f.slepian, flm=g.coefficients)
        if hasattr(f, "slepian")
        else g.coefficients
    )
    coefficients = f.convolve(g_coefficients, coefficients, shannon=shannon)

    filename += f"_convolved_{g.name}"
    return coefficients, filename


def main() -> None:
    args = read_args()

    mask = sleplet._mask_methods.create_default_region() if args.region else None

    f = sleplet._class_lists.COEFFICIENTS[
        sleplet._string_methods.convert_classes_list_to_snake_case(
            sleplet._class_lists.COEFFICIENTS,
        ).index(args.function)
    ](
        args.bandlimit,
        extra_args=args.extra_args,
        region=mask,
        noise=args.noise if args.noise is not None else None,
        smoothing=args.smoothing if args.smoothing is not None else None,
    )

    g = (
        sleplet._class_lists.MAPS_LM[
            sleplet._string_methods.convert_classes_list_to_snake_case(
                sleplet._class_lists.MAPS_LM,
            ).index(args.convolve)
        ](args.bandlimit)
        if isinstance(args.convolve, str)
        else None
    )

    # custom amplitude for noisy plots
    amplitude = sleplet.plot_methods.compute_amplitude_for_noisy_sphere_plots(f)

    plot(
        f,
        g,
        alpha_pi_frac=args.alpha,
        beta_pi_frac=args.beta,
        gamma_pi_frac=args.gamma,
        annotations=args.outline,
        normalise=not args.unnormalise,
        method=args.method,
        plot_type=args.type,
        upsample=not args.unzeropad,
        earth_view=args.perspective,
        amplitude=amplitude,
    )


if __name__ == "__main__":
    main()



"""Contains the abstract `Coefficients` class."""
import abc
import dataclasses

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._convolution_methods
import sleplet._mask_methods
import sleplet._string_methods
import sleplet._validation
import sleplet.slepian.region

_COEFFICIENTS_TO_NOT_MASK: set[str] = {"slepian", "south", "america"}


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class Coefficients:
    """
    Abstract parent class to handle harmonic/Slepian coefficients on the
    sphere.
    """

    L: int
    """The spherical harmonic bandlimit."""
    _: dataclasses.KW_ONLY
    extra_args: list[int] | None = None
    """Control the extra arguments for the given set of spherical
    coefficients. Only to be set by the `sphere` CLI."""
    noise: float | None = None
    """How much to noise the data."""
    region: sleplet.slepian.region.Region | None = None
    """Whether to set a region or not, used in the Slepian case."""
    smoothing: int | None = None
    """How much to smooth the topographic map of the Earth by."""
    _unnoised_coefficients: (
        npt.NDArray[np.complex_ | np.float_] | None
    ) = pydantic.Field(
        default=None,
        init_var=False,
        repr=False,
    )
    coefficients: npt.NDArray[np.complex_ | np.float_] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )
    name: str = pydantic.Field(default="", init_var=False, repr=False)
    reality: bool = pydantic.Field(default=False, init_var=False, repr=False)
    snr: float | None = pydantic.Field(default=None, init_var=False, repr=False)
    spin: int = pydantic.Field(default=0, init_var=False, repr=False)
    wavelet_coefficients: npt.NDArray[np.complex_ | np.float_] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )
    wavelets: npt.NDArray[np.complex_ | np.float_] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )

    def __post_init__(self: typing_extensions.Self) -> None:
        self._setup_args()
        self.name = self._create_name()
        self.spin = self._set_spin()
        self.reality = self._set_reality()
        self.coefficients = self._create_coefficients()
        self._add_details_to_name()
        self._unnoised_coefficients, self.snr = self._add_noise_to_signal()

    def translate(
        self: typing_extensions.Self,
        alpha: float,
        beta: float,
        *,
        shannon: int | None = None,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        r"""
        Perform the translation of the coefficients, used in the sifting convolution.

        Args:
            alpha: The point on the 2-sphere to translate to, i.e. the \(\phi\) value.
            beta: The point on the 2-sphere to translate to, i.e. the \(\theta\) value.
            shannon: The Shannon number, only used in the Slepian case.

        Returns:
            The translated spherical harmonic coefficients.
        """
        g_coefficients = self._translation_helper(alpha, beta)
        return (
            g_coefficients
            if "dirac_delta" in self.name
            else self.convolve(self.coefficients, g_coefficients, shannon=shannon)
        )

    def convolve(
        self: typing_extensions.Self,
        f_coefficient: npt.NDArray[np.complex_ | np.float_],
        g_coefficient: npt.NDArray[np.complex_ | np.float_],
        *,
        shannon: int | None = None,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        """
        Perform the sifting convolution of the two inputs.

        Args:
            f_coefficient: Input harmonic/Slepian coefficients.
            g_coefficient: Input harmonic/Slepian coefficients.
            shannon: The Shannon number, only used in the Slepian case.

        Returns:
            The sifting convolution of the two inputs.
        """
        # translation/convolution are not real for general function
        self.reality = False
        return sleplet._convolution_methods.sifting_convolution(
            f_coefficient,
            g_coefficient,
            shannon=shannon,
        )

    def _add_details_to_name(self: typing_extensions.Self) -> None:
        """
        Add region to the name if present if not a Slepian function
        adds noise/smoothing if appropriate and bandlimit.
        """
        if (
            isinstance(self.region, sleplet.slepian.region.Region)
            and not set(self.name.split("_")) & _COEFFICIENTS_TO_NOT_MASK
        ):
            self.name += f"_{self.region._name_ending}"
        if self.noise is not None:
            self.name += f"{sleplet._string_methods.filename_args(self.noise, 'noise')}"
        if self.smoothing is not None:
            self.name += (
                f"{sleplet._string_methods.filename_args(self.smoothing, 'smoothed')}"
            )
        self.name += f"_L{self.L}"

    @abc.abstractmethod
    def rotate(
        self: typing_extensions.Self,
        alpha: float,
        beta: float,
        *,
        gamma: float = 0,
    ) -> npt.NDArray[np.complex_]:
        r"""
        Rotates given flm on the sphere by alpha/beta/gamma.

        Args:
            alpha: The third Euler angle, a \(\alpha\) rotation about the z-axis.
            beta: The second Euler angle, a \(\beta\) rotation about the y-axis.
            gamma: The first Euler angle, a \(\gamma\) rotation about the z-axis.

        Returns:
            The rotated spherical harmonic coefficients.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _translation_helper(
        self: typing_extensions.Self,
        alpha: float,
        beta: float,
    ) -> npt.NDArray[np.complex_]:
        """Compute the basis function at omega' for translation."""
        raise NotImplementedError

    @abc.abstractmethod
    def _add_noise_to_signal(
        self: typing_extensions.Self,
    ) -> tuple[npt.NDArray[np.complex_ | np.float_] | None, float | None]:
        """Add Gaussian white noise to the signal."""
        raise NotImplementedError

    @abc.abstractmethod
    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        """Create the flm on the north pole."""
        raise NotImplementedError

    @abc.abstractmethod
    def _create_name(self: typing_extensions.Self) -> str:
        """Create the name of the function."""
        raise NotImplementedError

    @abc.abstractmethod
    def _set_reality(self: typing_extensions.Self) -> bool:
        """Set the reality flag to speed up computations."""
        raise NotImplementedError

    @abc.abstractmethod
    def _set_spin(self: typing_extensions.Self) -> int:
        """Set the spin value in computations."""
        raise NotImplementedError

    @abc.abstractmethod
    def _setup_args(self: typing_extensions.Self) -> None:
        """Initialise function specific args either default value or user input."""
        raise NotImplementedError
